=== modified file 'NEWS'
--- NEWS	2008-09-23 16:14:54 +0000
+++ NEWS	2008-09-24 12:57:29 +0000
@@ -4,6 +4,21 @@
 
 .. contents::
 
+bzr 1.7.1rc1 2008-09-24
+-----------------------
+
+This release just includes an update to how the merge algorithm handles
+file paths when we encounter complex history.
+
+  FEATURES:
+
+    * If we encounter a criss-cross in history, use information from
+      direct Least Common Ancestors to resolve inventory shape (locations
+      of files, adds, deletes, etc). This is similar in concept to using
+      ``--lca`` for merging file texts, only applied to paths.
+      (John Arbash Meinel)
+
+
 bzr 1.7 2008-09-23
 ------------------
 

=== modified file 'bzr'
--- bzr	2008-08-29 19:32:00 +0000
+++ bzr	2008-09-24 12:58:07 +0000
@@ -22,7 +22,7 @@
 import sys
 
 # update this on each release
-_script_version = (1, 7, 0)
+_script_version = (1, 7, 1)
 
 if __doc__ is None:
     print "bzr does not support python -OO."

=== modified file 'bzrlib/__init__.py'
--- bzrlib/__init__.py	2008-09-23 16:14:54 +0000
+++ bzrlib/__init__.py	2008-09-24 12:58:07 +0000
@@ -41,7 +41,7 @@
 # Python version 2.0 is (2, 0, 0, 'final', 0)."  Additionally we use a
 # releaselevel of 'dev' for unreleased under-development code.
 
-version_info = (1, 7, 0, 'final', 0)
+version_info = (1, 7, 1, 'candidate', 1)
 
 
 # API compatibility version: bzrlib is currently API compatible with 1.7.

=== modified file 'bzrlib/branchbuilder.py'
--- bzrlib/branchbuilder.py	2008-07-29 16:16:36 +0000
+++ bzrlib/branchbuilder.py	2008-07-29 16:34:09 +0000
@@ -129,7 +129,7 @@
             ('add', ('path', 'file-id', 'kind', 'content' or None))
             ('modify', ('file-id', 'new-content'))
             ('unversion', 'file-id')
-            # not supported yet: ('rename', ('orig-path', 'new-path'))
+            ('rename', ('orig-path', 'new-path'))
         :param message: An optional commit message, if not supplied, a default
             commit message will be written.
         :return: The revision_id of the new commit
@@ -157,10 +157,7 @@
             to_add_kinds = []
             new_contents = {}
             to_unversion_ids = []
-            # TODO: MemoryTree doesn't support rename() or
-            #       apply_inventory_delta, so we'll postpone allowing renames
-            #       for now
-            # to_rename = []
+            to_rename = []
             for action, info in actions:
                 if action == 'add':
                     path, file_id, kind, content = info
@@ -177,6 +174,9 @@
                     new_contents[file_id] = content
                 elif action == 'unversion':
                     to_unversion_ids.append(info)
+                elif action == 'rename':
+                    from_relpath, to_relpath = info
+                    to_rename.append((from_relpath, to_relpath))
                 else:
                     raise ValueError('Unknown build action: "%s"' % (action,))
             if to_unversion_ids:
@@ -187,6 +187,8 @@
                     tree.add([path], [file_id], ['directory'])
                 else:
                     tree.mkdir(path, file_id)
+            for from_relpath, to_relpath in to_rename:
+                tree.rename_one(from_relpath, to_relpath)
             tree.add(to_add_files, to_add_file_ids, to_add_kinds)
             for file_id, content in new_contents.iteritems():
                 tree.put_file_bytes_non_atomic(file_id, content)

=== modified file 'bzrlib/memorytree.py'
--- bzrlib/memorytree.py	2008-09-02 00:29:28 +0000
+++ bzrlib/memorytree.py	2008-09-24 12:51:22 +0000
@@ -21,10 +21,12 @@
 
 
 from copy import deepcopy
+import os
 
 from bzrlib import (
     errors,
     mutabletree,
+    osutils,
     revision as _mod_revision,
     )
 from bzrlib.decorators import needs_read_lock, needs_write_lock
@@ -99,6 +101,14 @@
             return None, False, None
         return entry.kind, entry.executable, None
 
+    @needs_tree_write_lock
+    def rename_one(self, from_rel, to_rel):
+        file_id = self.path2id(from_rel)
+        to_dir, to_tail = os.path.split(to_rel)
+        to_parent_id = self.path2id(to_dir)
+        self._file_transport.move(from_rel, to_rel)
+        self._inventory.rename(file_id, to_parent_id, to_tail)
+
     def path_content_summary(self, path):
         """See Tree.path_content_summary."""
         id = self.path2id(path)

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2008-09-05 05:12:35 +0000
+++ bzrlib/merge.py	2008-09-24 12:53:22 +0000
@@ -28,6 +28,7 @@
     patiencediff,
     registry,
     revision as _mod_revision,
+    tree as _mod_tree,
     tsort,
     )
 from bzrlib.branch import Branch
@@ -96,6 +97,8 @@
         self._revision_graph = revision_graph
         self._base_is_ancestor = None
         self._base_is_other_ancestor = None
+        self._is_criss_cross = None
+        self._lca_trees = None
 
     @property
     def revision_graph(self):
@@ -353,15 +356,45 @@
                      ensure_null(self.other_basis)]
         if NULL_REVISION in revisions:
             self.base_rev_id = NULL_REVISION
+            self.base_tree = self.revision_tree(self.base_rev_id)
+            self._is_criss_cross = False
         else:
-            self.base_rev_id, steps = self.revision_graph.find_unique_lca(
-                revisions[0], revisions[1], count_steps=True)
+            lcas = self.revision_graph.find_lca(revisions[0], revisions[1])
+            self._is_criss_cross = False
+            if len(lcas) == 0:
+                self.base_rev_id = NULL_REVISION
+            elif len(lcas) == 1:
+                self.base_rev_id = list(lcas)[0]
+            else: # len(lcas) > 1
+                if len(lcas) > 2:
+                    # find_unique_lca can only handle 2 nodes, so we have to
+                    # start back at the beginning. It is a shame to traverse
+                    # the graph again, but better than re-implementing
+                    # find_unique_lca.
+                    self.base_rev_id = self.revision_graph.find_unique_lca(
+                                            revisions[0], revisions[1])
+                else:
+                    self.base_rev_id = self.revision_graph.find_unique_lca(
+                                            *lcas)
+                self._is_criss_cross = True
             if self.base_rev_id == NULL_REVISION:
                 raise UnrelatedBranches()
-            if steps > 1:
+            if self._is_criss_cross:
                 warning('Warning: criss-cross merge encountered.  See bzr'
                         ' help criss-cross.')
-        self.base_tree = self.revision_tree(self.base_rev_id)
+                interesting_revision_ids = [self.base_rev_id]
+                interesting_revision_ids.extend(lcas)
+                interesting_trees = dict((t.get_revision_id(), t)
+                    for t in self.this_branch.repository.revision_trees(
+                        interesting_revision_ids))
+                self._cached_trees.update(interesting_trees)
+                self.base_tree = interesting_trees.pop(self.base_rev_id)
+                sorted_lca_keys = self.revision_graph.find_merge_order(
+                    revisions[0], lcas)
+                self._lca_trees = [interesting_trees[key]
+                                   for key in sorted_lca_keys]
+            else:
+                self.base_tree = self.revision_tree(self.base_rev_id)
         self.base_is_ancestor = True
         self.base_is_other_ancestor = True
 
@@ -409,6 +442,9 @@
         if self.merge_type.supports_cherrypick:
             kwargs['cherrypick'] = (not self.base_is_ancestor or
                                     not self.base_is_other_ancestor)
+        if self._is_criss_cross and getattr(self.merge_type,
+                                            'supports_lca_trees', False):
+            kwargs['lca_trees'] = self._lca_trees
         return self.merge_type(pb=self._pb,
                                change_reporter=self.change_reporter,
                                **kwargs)
@@ -460,6 +496,23 @@
         return len(merge.cooked_conflicts)
 
 
+class _InventoryNoneEntry(object):
+    """This represents an inventory entry which *isn't there*.
+
+    It simplifies the merging logic if we always have an InventoryEntry, even
+    if it isn't actually present
+    """
+    executable = None
+    kind = None
+    name = None
+    parent_id = None
+    revision = None
+    symlink_target = None
+    text_sha1 = None
+
+_none_entry = _InventoryNoneEntry()
+
+
 class Merge3Merger(object):
     """Three-way merger that uses the merge3 text merger"""
     requires_base = True
@@ -469,12 +522,13 @@
     supports_cherrypick = True
     supports_reverse_cherrypick = True
     winner_idx = {"this": 2, "other": 1, "conflict": 1}
+    supports_lca_trees = True
 
     def __init__(self, working_tree, this_tree, base_tree, other_tree, 
                  interesting_ids=None, reprocess=False, show_base=False,
                  pb=DummyProgress(), pp=None, change_reporter=None,
                  interesting_files=None, do_merge=True,
-                 cherrypick=False):
+                 cherrypick=False, lca_trees=None):
         """Initialize the merger object and perform the merge.
 
         :param working_tree: The working tree to apply the merge to
@@ -496,6 +550,9 @@
             be combined with interesting_ids.  If neither interesting_files nor
             interesting_ids is specified, all files may participate in the
             merge.
+        :param lca_trees: Can be set to a dictionary of {revision_id:rev_tree}
+            if the ancestry was found to include a criss-cross merge.
+            Otherwise should be None.
         """
         object.__init__(self)
         if interesting_files is not None and interesting_ids is not None:
@@ -510,6 +567,12 @@
         self.cooked_conflicts = []
         self.reprocess = reprocess
         self.show_base = show_base
+        self._lca_trees = lca_trees
+        # Uncommenting this will change the default algorithm to always use
+        # _entries_lca. This can be useful for running the test suite and
+        # making sure we haven't missed any corner cases.
+        # if lca_trees is None:
+        #     self._lca_trees = [self.base_tree]
         self.pb = pb
         self.pp = pp
         self.change_reporter = change_reporter
@@ -556,19 +619,24 @@
         return self.tt
 
     def _compute_transform(self):
-        entries = self._entries3()
+        if self._lca_trees is None:
+            entries = self._entries3()
+            resolver = self._three_way
+        else:
+            entries = self._entries_lca()
+            resolver = self._lca_multi_way
         child_pb = ui.ui_factory.nested_progress_bar()
         try:
             for num, (file_id, changed, parents3, names3,
                       executable3) in enumerate(entries):
                 child_pb.update('Preparing file merge', num, len(entries))
-                self._merge_names(file_id, parents3, names3)
+                self._merge_names(file_id, parents3, names3, resolver=resolver)
                 if changed:
                     file_status = self.merge_contents(file_id)
                 else:
                     file_status = 'unmodified'
                 self._merge_executable(file_id,
-                    executable3, file_status)
+                    executable3, file_status, resolver=resolver)
         finally:
             child_pb.finished()
         self.fix_root()
@@ -620,6 +688,177 @@
             result.append((file_id, changed, parents3, names3, executable3))
         return result
 
+    def _entries_lca(self):
+        """Gather data about files modified between multiple trees.
+
+        This compares OTHER versus all LCA trees, and for interesting entries,
+        it then compares with THIS and BASE.
+
+        For the multi-valued entries, the format will be (BASE, [lca1, lca2])
+        :return: [(file_id, changed, parents, names, executable)]
+            file_id     Simple file_id of the entry
+            changed     Boolean, True if the kind or contents changed
+                        else False
+            parents     ((base, [parent_id, in, lcas]), parent_id_other,
+                         parent_id_this)
+            names       ((base, [name, in, lcas]), name_in_other, name_in_this)
+            executable  ((base, [exec, in, lcas]), exec_in_other, exec_in_this)
+        """
+        if self.interesting_files is not None:
+            lookup_trees = [self.this_tree, self.base_tree]
+            lookup_trees.extend(self._lca_trees)
+            # I think we should include the lca trees as well
+            interesting_ids = self.other_tree.paths2ids(self.interesting_files,
+                                                        lookup_trees)
+        else:
+            interesting_ids = self.interesting_ids
+        result = []
+        walker = _mod_tree.MultiWalker(self.other_tree, self._lca_trees)
+
+        base_inventory = self.base_tree.inventory
+        this_inventory = self.this_tree.inventory
+        for path, file_id, other_ie, lca_values in walker.iter_all():
+            # Is this modified at all from any of the other trees?
+            if other_ie is None:
+                other_ie = _none_entry
+            if interesting_ids is not None and file_id not in interesting_ids:
+                continue
+
+            # If other_revision is found in any of the lcas, that means this
+            # node is uninteresting. This is because when merging, if there are
+            # multiple heads(), we have to create a new node. So if we didn't,
+            # we know that the ancestry is linear, and that OTHER did not
+            # modify anything
+            # See doc/developers/lca_merge_resolution.txt for details
+            other_revision = other_ie.revision
+            if other_revision is not None:
+                # We can't use this shortcut when other_revision is None,
+                # because it may be None because things are WorkingTrees, and
+                # not because it is *actually* None.
+                is_unmodified = False
+                for lca_path, ie in lca_values:
+                    if ie is not None and ie.revision == other_revision:
+                        is_unmodified = True
+                        break
+                if is_unmodified:
+                    continue
+
+            lca_entries = []
+            for lca_path, lca_ie in lca_values:
+                if lca_ie is None:
+                    lca_entries.append(_none_entry)
+                else:
+                    lca_entries.append(lca_ie)
+
+            if file_id in base_inventory:
+                base_ie = base_inventory[file_id]
+            else:
+                base_ie = _none_entry
+
+            if file_id in this_inventory:
+                this_ie = this_inventory[file_id]
+            else:
+                this_ie = _none_entry
+
+            lca_kinds = []
+            lca_parent_ids = []
+            lca_names = []
+            lca_executable = []
+            for lca_ie in lca_entries:
+                lca_kinds.append(lca_ie.kind)
+                lca_parent_ids.append(lca_ie.parent_id)
+                lca_names.append(lca_ie.name)
+                lca_executable.append(lca_ie.executable)
+
+            kind_winner = self._lca_multi_way(
+                (base_ie.kind, lca_kinds),
+                other_ie.kind, this_ie.kind)
+            parent_id_winner = self._lca_multi_way(
+                (base_ie.parent_id, lca_parent_ids),
+                other_ie.parent_id, this_ie.parent_id)
+            name_winner = self._lca_multi_way(
+                (base_ie.name, lca_names),
+                other_ie.name, this_ie.name)
+
+            content_changed = True
+            if kind_winner == 'this':
+                # No kind change in OTHER, see if there are *any* changes
+                if other_ie.kind == None:
+                    # No content and 'this' wins the kind, so skip this?
+                    # continue
+                    pass
+                elif other_ie.kind == 'directory':
+                    if parent_id_winner == 'this' and name_winner == 'this':
+                        # No change for this directory in OTHER, skip
+                        continue
+                    content_changed = False
+                elif other_ie.kind == 'file':
+                    def get_sha1(ie, tree):
+                        if ie.kind != 'file':
+                            return None
+                        return tree.get_file_sha1(file_id)
+                    base_sha1 = get_sha1(base_ie, self.base_tree)
+                    lca_sha1s = [get_sha1(ie, tree) for ie, tree
+                                 in zip(lca_entries, self._lca_trees)]
+                    this_sha1 = get_sha1(this_ie, self.this_tree)
+                    other_sha1 = get_sha1(other_ie, self.other_tree)
+                    sha1_winner = self._lca_multi_way(
+                        (base_sha1, lca_sha1s), other_sha1, this_sha1,
+                        allow_overriding_lca=False)
+                    exec_winner = self._lca_multi_way(
+                        (base_ie.executable, lca_executable),
+                        other_ie.executable, this_ie.executable)
+                    if (parent_id_winner == 'this' and name_winner == 'this'
+                        and sha1_winner == 'this' and exec_winner == 'this'):
+                        # No kind, parent, name, exec, or content change for
+                        # OTHER, so this node is not considered interesting
+                        continue
+                    if sha1_winner == 'this':
+                        content_changed = False
+                elif other_ie.kind == 'symlink':
+                    def get_target(ie, tree):
+                        if ie.kind != 'symlink':
+                            return None
+                        return tree.get_symlink_target(file_id)
+                    base_target = get_target(base_ie, self.base_tree)
+                    lca_targets = [get_target(ie, tree) for ie, tree
+                                   in zip(lca_entries, self._lca_trees)]
+                    this_target = get_target(this_ie, self.this_tree)
+                    other_target = get_target(other_ie, self.other_tree)
+                    target_winner = self._lca_multi_way(
+                        (base_target, lca_targets),
+                        other_target, this_target)
+                    if (parent_id_winner == 'this' and name_winner == 'this'
+                        and target_winner == 'this'):
+                        # No kind, parent, name, or symlink target change
+                        # not interesting
+                        continue
+                    if target_winner == 'this':
+                        content_changed = False
+                elif other_ie.kind == 'tree-reference':
+                    # The 'changed' information seems to be handled at a higher
+                    # level. At least, _entries3 returns False for content
+                    # changed, even when at a new revision_id.
+                    content_changed = False
+                    if (parent_id_winner == 'this' and name_winner == 'this'):
+                        # Nothing interesting
+                        continue
+                else:
+                    raise AssertionError('unhandled kind: %s' % other_ie.kind)
+                # XXX: We need to handle kind == 'symlink'
+
+            # If we have gotten this far, that means something has changed
+            result.append((file_id, content_changed,
+                           ((base_ie.parent_id, lca_parent_ids),
+                            other_ie.parent_id, this_ie.parent_id),
+                           ((base_ie.name, lca_names),
+                            other_ie.name, this_ie.name),
+                           ((base_ie.executable, lca_executable),
+                            other_ie.executable, this_ie.executable)
+                          ))
+        return result
+
+
     def fix_root(self):
         try:
             self.tt.final_kind(self.tt.root)
@@ -714,6 +953,56 @@
             return "other"
 
     @staticmethod
+    def _lca_multi_way(bases, other, this, allow_overriding_lca=True):
+        """Consider LCAs when determining whether a change has occurred.
+
+        If LCAS are all identical, this is the same as a _three_way comparison.
+
+        :param bases: value in (BASE, [LCAS])
+        :param other: value in OTHER
+        :param this: value in THIS
+        :param allow_overriding_lca: If there is more than one unique lca
+            value, allow OTHER to override THIS if it has a new value, and
+            THIS only has an lca value, or vice versa. This is appropriate for
+            truly scalar values, not as much for non-scalars.
+        :return: 'this', 'other', or 'conflict' depending on whether an entry
+            changed or not.
+        """
+        # See doc/developers/lca_merge_resolution.txt for details about this
+        # algorithm.
+        if other == this:
+            # Either Ambiguously clean, or nothing was actually changed. We
+            # don't really care
+            return 'this'
+        base_val, lca_vals = bases
+        # Remove 'base_val' from the lca_vals, because it is not interesting
+        filtered_lca_vals = [lca_val for lca_val in lca_vals
+                                      if lca_val != base_val]
+        if len(filtered_lca_vals) == 0:
+            return Merge3Merger._three_way(base_val, other, this)
+
+        unique_lca_vals = set(filtered_lca_vals)
+        if len(unique_lca_vals) == 1:
+            return Merge3Merger._three_way(unique_lca_vals.pop(), other, this)
+
+        if allow_overriding_lca:
+            if other in unique_lca_vals:
+                if this in unique_lca_vals:
+                    # Each side picked a different lca, conflict
+                    return 'conflict'
+                else:
+                    # This has a value which supersedes both lca values, and
+                    # other only has an lca value
+                    return 'this'
+            elif this in unique_lca_vals:
+                # OTHER has a value which supersedes both lca values, and this
+                # only has an lca value
+                return 'other'
+
+        # At this point, the lcas disagree, and the tips disagree
+        return 'conflict'
+
+    @staticmethod
     def scalar_three_way(this_tree, base_tree, other_tree, file_id, key):
         """Do a three-way test on a scalar.
         Return "this", "other" or "conflict", depending whether a value wins.
@@ -751,16 +1040,17 @@
             else:
                 names.append(entry.name)
                 parents.append(entry.parent_id)
-        return self._merge_names(file_id, parents, names)
+        return self._merge_names(file_id, parents, names,
+                                 resolver=self._three_way)
 
-    def _merge_names(self, file_id, parents, names):
+    def _merge_names(self, file_id, parents, names, resolver):
         """Perform a merge on file_id names and parents"""
         base_name, other_name, this_name = names
         base_parent, other_parent, this_parent = parents
 
-        name_winner = self._three_way(*names)
+        name_winner = resolver(*names)
 
-        parent_id_winner = self._three_way(*parents)
+        parent_id_winner = resolver(*parents)
         if this_name is None:
             if name_winner == "this":
                 name_winner = "other"
@@ -954,14 +1244,16 @@
         """Perform a merge on the execute bit."""
         executable = [self.executable(t, file_id) for t in (self.base_tree,
                       self.other_tree, self.this_tree)]
-        self._merge_executable(file_id, executable, file_status)
+        self._merge_executable(file_id, executable, file_status,
+                               resolver=self._three_way)
 
-    def _merge_executable(self, file_id, executable, file_status):
+    def _merge_executable(self, file_id, executable, file_status,
+                          resolver):
         """Perform a merge on the execute bit."""
         base_executable, other_executable, this_executable = executable
         if file_status == "deleted":
             return
-        winner = self._three_way(*executable)
+        winner = resolver(*executable)
         if winner == "conflict":
         # There must be a None in here, if we have a conflict, but we
         # need executability since file status was not deleted.

=== modified file 'bzrlib/tests/test_branchbuilder.py'
--- bzrlib/tests/test_branchbuilder.py	2008-07-29 16:16:36 +0000
+++ bzrlib/tests/test_branchbuilder.py	2008-07-29 16:34:09 +0000
@@ -197,8 +197,23 @@
             builder.build_snapshot, 'B-id', None, [('weirdo', ('foo',))])
         self.assertEqual('Unknown build action: "weirdo"', str(e))
 
-    # TODO: rename a file/directory, but rename isn't supported by the
-    #       MemoryTree api yet, so for now we wait until it is used
+    def test_rename(self):
+        builder = self.build_a_rev()
+        builder.build_snapshot('B-id', None,
+            [('rename', ('a', 'b'))])
+        rev_tree = builder.get_branch().repository.revision_tree('B-id')
+        self.assertTreeShape([(u'', 'a-root-id', 'directory'),
+                              (u'b', 'a-id', 'file')], rev_tree)
+
+    def test_rename_into_subdir(self):
+        builder = self.build_a_rev()
+        builder.build_snapshot('B-id', None,
+            [('add', ('dir', 'dir-id', 'directory', None)),
+             ('rename', ('a', 'dir/a'))])
+        rev_tree = builder.get_branch().repository.revision_tree('B-id')
+        self.assertTreeShape([(u'', 'a-root-id', 'directory'),
+                              (u'dir', 'dir-id', 'directory'),
+                              (u'dir/a', 'a-id', 'file')], rev_tree)
 
     def test_set_parent(self):
         builder = self.build_a_rev()

=== modified file 'bzrlib/tests/test_memorytree.py'
--- bzrlib/tests/test_memorytree.py	2008-07-22 18:12:25 +0000
+++ bzrlib/tests/test_memorytree.py	2008-07-23 02:35:09 +0000
@@ -182,3 +182,42 @@
         rev_id = tree.commit('first post')
         tree.unlock()
         self.assertEqual(rev_id, tree.last_revision())
+
+    def test_rename_file(self):
+        tree = self.make_branch_and_memory_tree('branch')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        tree.add(['', 'foo'], ['root-id', 'foo-id'], ['directory', 'file'])
+        tree.put_file_bytes_non_atomic('foo-id', 'content\n')
+        tree.commit('one', rev_id='rev-one')
+        tree.rename_one('foo', 'bar')
+        self.assertEqual('bar', tree.id2path('foo-id'))
+        self.assertEqual('content\n', tree._file_transport.get_bytes('bar'))
+        self.assertRaises(errors.NoSuchFile,
+                          tree._file_transport.get_bytes, 'foo')
+        tree.commit('two', rev_id='rev-two')
+        self.assertEqual('content\n', tree._file_transport.get_bytes('bar'))
+        self.assertRaises(errors.NoSuchFile,
+                          tree._file_transport.get_bytes, 'foo')
+
+        rev_tree2 = tree.branch.repository.revision_tree('rev-two')
+        self.assertEqual('bar', rev_tree2.id2path('foo-id'))
+        self.assertEqual('content\n', rev_tree2.get_file_text('foo-id'))
+
+    def test_rename_file_to_subdir(self):
+        tree = self.make_branch_and_memory_tree('branch')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        tree.add('')
+        tree.mkdir('subdir', 'subdir-id')
+        tree.add('foo', 'foo-id', 'file')
+        tree.put_file_bytes_non_atomic('foo-id', 'content\n')
+        tree.commit('one', rev_id='rev-one')
+
+        tree.rename_one('foo', 'subdir/bar')
+        self.assertEqual('subdir/bar', tree.id2path('foo-id'))
+        self.assertEqual('content\n',
+                         tree._file_transport.get_bytes('subdir/bar'))
+        tree.commit('two', rev_id='rev-two')
+        rev_tree2 = tree.branch.repository.revision_tree('rev-two')
+        self.assertEqual('subdir/bar', rev_tree2.id2path('foo-id'))

=== modified file 'bzrlib/tests/test_merge.py'
--- bzrlib/tests/test_merge.py	2008-08-27 20:32:22 +0000
+++ bzrlib/tests/test_merge.py	2008-09-05 03:11:40 +0000
@@ -21,9 +21,11 @@
     conflicts,
     errors,
     knit,
+    memorytree,
     merge as _mod_merge,
     option,
     progress,
+    tests,
     transform,
     versionedfile,
     )
@@ -1107,3 +1109,1638 @@
     def test_merge_move_and_change(self):
         self.expectFailure("lca merge doesn't conflict for move and change",
             super(TestLCAMerge, self).test_merge_move_and_change)
+
+
+class LoggingMerger(object):
+    # These seem to be the required attributes
+    requires_base = False
+    supports_reprocess = False
+    supports_show_base = False
+    supports_cherrypick = False
+    # We intentionally do not define supports_lca_trees
+
+    def __init__(self, *args, **kwargs):
+        self.args = args
+        self.kwargs = kwargs
+
+
+class TestMergerBase(TestCaseWithMemoryTransport):
+    """Common functionality for Merger tests that don't write to disk."""
+
+    def get_builder(self):
+        builder = self.make_branch_builder('path')
+        builder.start_series()
+        self.addCleanup(builder.finish_series)
+        return builder
+
+    def setup_simple_graph(self):
+        """Create a simple 3-node graph.
+
+        :return: A BranchBuilder
+        """
+        #
+        #  A
+        #  |\
+        #  B C
+        #
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', ('', None, 'directory', None))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        return builder
+
+    def setup_criss_cross_graph(self):
+        """Create a 5-node graph with a criss-cross.
+
+        :return: A BranchBuilder
+        """
+        # A
+        # |\
+        # B C
+        # |X|
+        # D E
+        builder = self.setup_simple_graph()
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        return builder
+
+    def make_Merger(self, builder, other_revision_id,
+                    interesting_files=None, interesting_ids=None):
+        """Make a Merger object from a branch builder"""
+        mem_tree = memorytree.MemoryTree.create_on_branch(builder.get_branch())
+        mem_tree.lock_write()
+        self.addCleanup(mem_tree.unlock)
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            mem_tree, other_revision_id)
+        merger.set_interesting_files(interesting_files)
+        # It seems there is no matching function for set_interesting_ids
+        merger.interesting_ids = interesting_ids
+        merger.merge_type = _mod_merge.Merge3Merger
+        return merger
+
+
+class TestMergerInMemory(TestMergerBase):
+
+    def test_find_base(self):
+        merger = self.make_Merger(self.setup_simple_graph(), 'C-id')
+        self.assertEqual('A-id', merger.base_rev_id)
+        self.assertFalse(merger._is_criss_cross)
+        self.assertIs(None, merger._lca_trees)
+
+    def test_find_base_criss_cross(self):
+        builder = self.setup_criss_cross_graph()
+        merger = self.make_Merger(builder, 'E-id')
+        self.assertEqual('A-id', merger.base_rev_id)
+        self.assertTrue(merger._is_criss_cross)
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merger._lca_trees])
+        # If we swap the order, we should get a different lca order
+        builder.build_snapshot('F-id', ['E-id'], [])
+        merger = self.make_Merger(builder, 'D-id')
+        self.assertEqual(['C-id', 'B-id'], [t.get_revision_id()
+                                            for t in merger._lca_trees])
+
+    def test_find_base_triple_criss_cross(self):
+        #       A-.
+        #      / \ \
+        #     B   C F # F is merged into both branches
+        #     |\ /| |
+        #     | X | |\
+        #     |/ \| | :
+        #   : D   E |
+        #    \|   |/
+        #     G   H
+        builder = self.setup_criss_cross_graph()
+        builder.build_snapshot('F-id', ['A-id'], [])
+        builder.build_snapshot('H-id', ['E-id', 'F-id'], [])
+        builder.build_snapshot('G-id', ['D-id', 'F-id'], [])
+        merger = self.make_Merger(builder, 'H-id')
+        self.assertEqual(['B-id', 'C-id', 'F-id'],
+                         [t.get_revision_id() for t in merger._lca_trees])
+
+    def test_no_criss_cross_passed_to_merge_type(self):
+        class LCATreesMerger(LoggingMerger):
+            supports_lca_trees = True
+
+        merger = self.make_Merger(self.setup_simple_graph(), 'C-id')
+        merger.merge_type = LCATreesMerger
+        merge_obj = merger.make_merger()
+        self.assertIsInstance(merge_obj, LCATreesMerger)
+        self.assertFalse('lca_trees' in merge_obj.kwargs)
+
+    def test_criss_cross_passed_to_merge_type(self):
+        merger = self.make_Merger(self.setup_criss_cross_graph(), 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merger._lca_trees])
+
+    def test_criss_cross_not_supported_merge_type(self):
+        merger = self.make_Merger(self.setup_criss_cross_graph(), 'E-id')
+        # We explicitly do not define supports_lca_trees
+        merger.merge_type = LoggingMerger
+        merge_obj = merger.make_merger()
+        self.assertIsInstance(merge_obj, LoggingMerger)
+        self.assertFalse('lca_trees' in merge_obj.kwargs)
+
+    def test_criss_cross_unsupported_merge_type(self):
+        class UnsupportedLCATreesMerger(LoggingMerger):
+            supports_lca_trees = False
+
+        merger = self.make_Merger(self.setup_criss_cross_graph(), 'E-id')
+        merger.merge_type = UnsupportedLCATreesMerger
+        merge_obj = merger.make_merger()
+        self.assertIsInstance(merge_obj, UnsupportedLCATreesMerger)
+        self.assertFalse('lca_trees' in merge_obj.kwargs)
+
+
+class TestMergerEntriesLCA(TestMergerBase):
+
+    def make_merge_obj(self, builder, other_revision_id,
+                       interesting_files=None, interesting_ids=None):
+        merger = self.make_Merger(builder, other_revision_id,
+            interesting_files=interesting_files,
+            interesting_ids=interesting_ids)
+        return merger.make_merger()
+
+    def test_simple(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('a-id', 'a\nb\nC\nc\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('a-id', 'a\nB\nb\nc\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'a\nB\nb\nC\nc\nE\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('modify', ('a-id', 'a\nB\nb\nC\nc\n'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merge_obj._lca_trees])
+        self.assertEqual('A-id', merge_obj.base_tree.get_revision_id())
+        entries = list(merge_obj._entries_lca())
+
+        # (file_id, changed, parents, names, executable)
+        # BASE, lca1, lca2, OTHER, THIS
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'a', [u'a', u'a']), u'a', u'a'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_not_in_base(self):
+        # LCAs all have the same last-modified revision for the file, as do
+        # the tips, but the base has something different
+        #       A    base, doesn't have the file
+        #       |\
+        #       B C  B introduces 'foo', C introduces 'bar'
+        #       |X|
+        #       D E  D and E now both have 'foo' and 'bar'
+        #       |X|
+        #       F G  the files are now in F, G, D and E, but not in A
+        #            G modifies 'bar'
+
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('add', (u'bar', 'bar-id', 'file', 'd\ne\nf\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('add', (u'bar', 'bar-id', 'file', 'd\ne\nf\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('G-id', ['E-id', 'D-id'],
+            [('modify', (u'bar-id', 'd\ne\nf\nG\n'))])
+        builder.build_snapshot('F-id', ['D-id', 'E-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'G-id')
+
+        self.assertEqual(['D-id', 'E-id'], [t.get_revision_id()
+                                            for t in merge_obj._lca_trees])
+        self.assertEqual('A-id', merge_obj.base_tree.get_revision_id())
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('bar-id', True,
+                           ((None, [root_id, root_id]), root_id, root_id),
+                           ((None, [u'bar', u'bar']), u'bar', u'bar'),
+                           ((None, [False, False]), False, False)),
+                         ], entries)
+
+    def test_not_in_this(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('a-id', 'a\nB\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('a-id', 'a\nb\nC\nc\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'a\nB\nb\nC\nc\nE\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('unversion', 'a-id')])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merge_obj._lca_trees])
+        self.assertEqual('A-id', merge_obj.base_tree.get_revision_id())
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((root_id, [root_id, root_id]), root_id, None),
+                           ((u'a', [u'a', u'a']), u'a', None),
+                           ((False, [False, False]), False, None)),
+                         ], entries)
+
+    def test_file_not_in_one_lca(self):
+        #   A   # just root
+        #   |\
+        #   B C # B no file, C introduces a file
+        #   |X|
+        #   D E # D and E both have the file, unchanged from C
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], []) # Inherited from C
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], # Merged from C
+            [('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        self.assertEqual(['B-id', 'C-id'], [t.get_revision_id()
+                                            for t in merge_obj._lca_trees])
+        self.assertEqual('A-id', merge_obj.base_tree.get_revision_id())
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([], entries)
+
+    def test_not_in_other(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('unversion', 'a-id')])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((root_id, [root_id, root_id]), None, root_id),
+                           ((u'a', [u'a', u'a']), None, u'a'),
+                           ((False, [False, False]), None, False)),
+                         ], entries)
+
+    def test_not_in_other_or_lca(self):
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B nothing, C deletes foo
+        #       |X|
+        #       D E  D restores foo (same as B), E leaves it deleted
+        # We should emit an entry for this
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('unversion', 'foo-id')])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('foo-id', True,
+                           ((root_id, [root_id, None]), None, root_id),
+                           ((u'foo', [u'foo', None]), None, 'foo'),
+                           ((False, [False, None]), None, False)),
+                         ], entries)
+
+    def test_only_in_one_lca(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('unversion', 'a-id')])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((None, [None, root_id]), None, None),
+                           ((None, [None, u'a']), None, None),
+                           ((None, [None, False]), None, None)),
+                         ], entries)
+
+    def test_only_in_other(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', True,
+                           ((None, [None, None]), root_id, None),
+                           ((None, [None, None]), u'a', None),
+                           ((None, [None, None]), False, None)),
+                         ], entries)
+
+    def test_one_lca_supersedes(self):
+        # One LCA supersedes the other LCAs last modified value, but the
+        # value is not the same as BASE.
+        #       A    base, introduces 'foo', last mod A
+        #       |\
+        #       B C  B modifies 'foo' (mod B), C does nothing (mod A)
+        #       |X|
+        #       D E  D does nothing (mod B), E updates 'foo' (mod E)
+        #       |X|
+        #       F G  F updates 'foo' (mod F). G does nothing (mod E)
+        #
+        #   At this point, G should not be considered to modify 'foo', even
+        #   though its LCAs disagree. This is because the modification in E
+        #   completely supersedes the value in D.
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('foo-id', 'E content\n'))])
+        builder.build_snapshot('G-id', ['E-id', 'D-id'], [])
+        builder.build_snapshot('F-id', ['D-id', 'E-id'],
+            [('modify', ('foo-id', 'F content\n'))])
+        merge_obj = self.make_merge_obj(builder, 'G-id')
+
+        self.assertEqual([], list(merge_obj._entries_lca()))
+
+    def test_one_lca_supersedes_path(self):
+        # Double-criss-cross merge, the ultimate base value is different from
+        # the intermediate.
+        #   A    value 'foo'
+        #   |\
+        #   B C  B value 'bar', C = 'foo'
+        #   |X|
+        #   D E  D = 'bar', E supersedes to 'bing'
+        #   |X|
+        #   F G  F = 'bing', G supersedes to 'barry'
+        #
+        # In this case, we technically should not care about the value 'bar' for
+        # D, because it was clearly superseded by E's 'bing'. The
+        # per-file/attribute graph would actually look like:
+        #   A
+        #   |
+        #   B
+        #   |
+        #   E
+        #   |
+        #   G
+        #
+        # Because the other side of the merge never modifies the value, it just
+        # takes the value from the merge.
+        #
+        # ATM this fails because we will prune 'foo' from the LCAs, but we
+        # won't prune 'bar'. This is getting far off into edge-case land, so we
+        # aren't supporting it yet.
+        #
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('rename', ('foo', 'bing'))]) # override to bing
+        builder.build_snapshot('G-id', ['E-id', 'D-id'],
+            [('rename', ('bing', 'barry'))]) # override to barry
+        builder.build_snapshot('F-id', ['D-id', 'E-id'],
+            [('rename', ('bar', 'bing'))]) # Merge in E's change
+        merge_obj = self.make_merge_obj(builder, 'G-id')
+
+        self.expectFailure("We don't do an actual heads() check on lca values,"
+            " or use the per-attribute graph",
+            self.assertEqual, [], list(merge_obj._entries_lca()))
+
+    def test_one_lca_accidentally_pruned(self):
+        # Another incorrect resolution from the same basic flaw:
+        #   A    value 'foo'
+        #   |\
+        #   B C  B value 'bar', C = 'foo'
+        #   |X|
+        #   D E  D = 'bar', E reverts to 'foo'
+        #   |X|
+        #   F G  F = 'bing', G switches to 'bar'
+        #
+        # 'bar' will not be seen as an interesting change, because 'foo' will
+        # be pruned from the LCAs, even though it was newly introduced by E
+        # (superseding B).
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('G-id', ['E-id', 'D-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('F-id', ['D-id', 'E-id'],
+            [('rename', ('bar', 'bing'))]) # should end up conflicting
+        merge_obj = self.make_merge_obj(builder, 'G-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.expectFailure("We prune values from BASE even when relevant.",
+            self.assertEqual,
+                [('foo-id', False,
+                  ((root_id, [root_id, root_id]), root_id, root_id),
+                  ((u'foo', [u'bar', u'foo']), u'bar', u'bing'),
+                  ((False, [False, False]), False, False)),
+                ], entries)
+
+    def test_both_sides_revert(self):
+        # Both sides of a criss-cross revert the text to the lca
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B modifies 'foo', C modifies 'foo'
+        #       |X|
+        #       D E  D reverts to B, E reverts to C
+        # This should conflict
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('foo-id', 'C content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('foo-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'foo', u'foo']), u'foo', u'foo'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_different_lca_resolve_one_side_updates_content(self):
+        # Both sides converge, but then one side updates the text.
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B modifies 'foo', C modifies 'foo'
+        #       |X|
+        #       D E  D reverts to B, E reverts to C
+        #       |
+        #       F    F updates to a new value
+        # We need to emit an entry for 'foo', because D & E differed on the
+        # merge resolution
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('foo-id', 'C content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('F-id', ['D-id'],
+            [('modify', ('foo-id', 'F content\n'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('foo-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'foo', u'foo']), u'foo', u'foo'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_same_lca_resolution_one_side_updates_content(self):
+        # Both sides converge, but then one side updates the text.
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B modifies 'foo', C modifies 'foo'
+        #       |X|
+        #       D E  D and E use C's value
+        #       |
+        #       F    F updates to a new value
+        # I think it is a bug that this conflicts, but we don't have a way to
+        # detect otherwise. And because of:
+        #   test_different_lca_resolve_one_side_updates_content
+        # We need to conflict.
+
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('foo-id', 'C content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('modify', ('foo-id', 'C content\n'))]) # Same as E
+        builder.build_snapshot('F-id', ['D-id'],
+            [('modify', ('foo-id', 'F content\n'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.expectFailure("We don't detect that LCA resolution was the"
+                           " same on both sides",
+            self.assertEqual, [], entries)
+
+    def test_only_path_changed(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('rename', (u'a', u'b'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        # The content was not changed, only the path
+        self.assertEqual([('a-id', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'a', [u'a', u'a']), u'b', u'a'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_kind_changed(self):
+        # Identical content, except 'D' changes a-id into a directory
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('unversion', 'a-id'),
+             ('add', (u'a', 'a-id', 'directory', None))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        # Only the kind was changed (content)
+        self.assertEqual([('a-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'a', [u'a', u'a']), u'a', u'a'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_this_changed_kind(self):
+        # Identical content, but THIS changes a file to a directory
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('unversion', 'a-id'),
+             ('add', (u'a', 'a-id', 'directory', None))])
+        merge_obj = self.make_merge_obj(builder, 'E-id')
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        # Only the kind was changed (content)
+        self.assertEqual([], entries)
+
+    def test_interesting_files(self):
+        # Two files modified, but we should filter one of them
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'b', 'b-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('b-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_files=['b'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('b-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'b', [u'b', u'b']), u'b', u'b'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_interesting_file_in_this(self):
+        # This renamed the file, but it should still match the entry in other
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'b', 'b-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('b-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('rename', ('b', 'c'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_files=['c'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('b-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'b', [u'b', u'b']), u'b', u'c'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_interesting_file_in_base(self):
+        # This renamed the file, but it should still match the entry in BASE
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'c', 'c-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('c', 'b'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('rename', ('c', 'b'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('c-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_files=['c'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('c-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'c', [u'b', u'b']), u'b', u'b'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_interesting_file_in_lca(self):
+        # This renamed the file, but it should still match the entry in LCA
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'b', 'b-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('b', 'c'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('b-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('rename', ('c', 'b'))])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_files=['c'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('b-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'b', [u'c', u'b']), u'b', u'b'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_interesting_ids(self):
+        # Two files modified, but we should filter one of them
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'content\n')),
+             ('add', (u'b', 'b-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('a-id', 'new-content\n')),
+             ('modify', ('b-id', 'new-content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        merge_obj = self.make_merge_obj(builder, 'E-id',
+                                        interesting_ids=['b-id'])
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('b-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'b', [u'b', u'b']), u'b', u'b'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+
+
+class TestMergerEntriesLCAOnDisk(tests.TestCaseWithTransport):
+
+    def get_builder(self):
+        builder = self.make_branch_builder('path')
+        builder.start_series()
+        self.addCleanup(builder.finish_series)
+        return builder
+
+    def get_wt_from_builder(self, builder):
+        """Get a real WorkingTree from the builder."""
+        the_branch = builder.get_branch()
+        wt = the_branch.bzrdir.create_workingtree()
+        # Note: This is a little bit ugly, but we are holding the branch
+        #       write-locked as part of the build process, and we would like to
+        #       maintain that. So we just force the WT to re-use the same
+        #       branch object.
+        wt._branch = the_branch
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        return wt
+
+    def do_merge(self, builder, other_revision_id):
+        wt = self.get_wt_from_builder(builder)
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, other_revision_id)
+        merger.merge_type = _mod_merge.Merge3Merger
+        return wt, merger.do_merge()
+
+    def test_simple_lca(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('modify', ('a-id', 'a\nb\nc\nd\ne\nf\n'))])
+        wt, conflicts = self.do_merge(builder, 'E-id')
+        self.assertEqual(0, conflicts)
+        # The merge should have simply update the contents of 'a'
+        self.assertEqual('a\nb\nc\nd\ne\nf\n', wt.get_file_text('a-id'))
+
+    def test_conflict_without_lca(self):
+        # This test would cause a merge conflict, unless we use the lca trees
+        # to determine the real ancestry
+        #   A       Path at 'foo'
+        #  / \
+        # B   C     Path renamed to 'bar' in B
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     Path at 'bar' in D and E
+        #     |
+        #     F     Path at 'baz' in F, which supersedes 'bar' and 'foo'
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], # merge the rename
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('F-id', ['E-id'],
+            [('rename', ('bar', 'baz'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.assertEqual(0, conflicts)
+        # The merge should simply recognize that the final rename takes
+        # precedence
+        self.assertEqual('baz', wt.id2path('foo-id'))
+
+    def test_other_deletes_lca_renames(self):
+        # This test would cause a merge conflict, unless we use the lca trees
+        # to determine the real ancestry
+        #   A       Path at 'foo'
+        #  / \
+        # B   C     Path renamed to 'bar' in B
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     Path at 'bar' in D and E
+        #     |
+        #     F     F deletes 'bar'
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], # merge the rename
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('F-id', ['E-id'],
+            [('unversion', 'foo-id')])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.assertEqual(0, conflicts)
+        self.assertRaises(errors.NoSuchId, wt.id2path, 'foo-id')
+
+    def test_executable_changes(self):
+        #   A       Path at 'foo'
+        #  / \
+        # B   C
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E
+        #     |
+        #     F     Executable bit changed
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        # Have to use a real WT, because BranchBuilder doesn't support exec bit
+        wt = self.get_wt_from_builder(builder)
+        tt = transform.TreeTransform(wt)
+        try:
+            tt.set_executability(True, tt.trans_id_tree_file_id('foo-id'))
+            tt.apply()
+        except:
+            tt.finalize()
+            raise
+        self.assertTrue(wt.is_executable('foo-id'))
+        wt.commit('F-id', rev_id='F-id')
+        # Reset to D, so that we can merge F
+        wt.set_parent_ids(['D-id'])
+        wt.branch.set_last_revision_info(3, 'D-id')
+        wt.revert()
+        self.assertFalse(wt.is_executable('foo-id'))
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='F-id')
+        self.assertEqual(0, conflicts)
+        self.assertTrue(wt.is_executable('foo-id'))
+
+    def test_create_symlink(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A
+        #  / \
+        # B   C
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E
+        #     |
+        #     F     Add a symlink 'foo' => 'bar'
+        # Have to use a real WT, because BranchBuilder and MemoryTree don't
+        # have symlink support
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        # Have to use a real WT, because BranchBuilder doesn't support exec bit
+        wt = self.get_wt_from_builder(builder)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        self.assertEqual('bar', wt.get_symlink_target('foo-id'))
+        wt.commit('add symlink', rev_id='F-id')
+        # Reset to D, so that we can merge F
+        wt.set_parent_ids(['D-id'])
+        wt.branch.set_last_revision_info(3, 'D-id')
+        wt.revert()
+        self.assertIs(None, wt.path2id('foo'))
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='F-id')
+        self.assertEqual(0, conflicts)
+        self.assertEqual('foo-id', wt.path2id('foo'))
+        self.assertEqual('bar', wt.get_symlink_target('foo-id'))
+
+    def test_both_sides_revert(self):
+        # Both sides of a criss-cross revert the text to the lca
+        #       A    base, introduces 'foo'
+        #       |\
+        #       B C  B modifies 'foo', C modifies 'foo'
+        #       |X|
+        #       D E  D reverts to B, E reverts to C
+        # This should conflict
+        # This must be done with a real WorkingTree, because normally their
+        # inventory contains "None" rather than a real sha1
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'A content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('modify', ('foo-id', 'C content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'], [])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'E-id')
+        self.assertEqual(1, conflicts)
+        self.assertEqualDiff('<<<<<<< TREE\n'
+                             'B content\n'
+                             '=======\n'
+                             'C content\n'
+                             '>>>>>>> MERGE-SOURCE\n',
+                             wt.get_file_text('foo-id'))
+
+    def test_modified_symlink(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A       Create symlink foo => bar
+        #  / \
+        # B   C     B relinks foo => baz
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     D & E have foo => baz
+        #     |
+        #     F     F changes it to bing
+        #
+        # Merging D & F should result in F cleanly overriding D, because D's
+        # value actually comes from B
+
+        # Have to use a real WT, because BranchBuilder and MemoryTree don't
+        # have symlink support
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('add symlink', rev_id='A-id')
+        os.remove('path/foo')
+        os.symlink('baz', 'path/foo')
+        wt.commit('foo => baz', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('baz', wt.get_symlink_target('foo-id'))
+        wt.commit('E merges C & B', rev_id='E-id')
+        os.remove('path/foo')
+        os.symlink('bing', 'path/foo')
+        wt.commit('F foo => bing', rev_id='F-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        wt.commit('D merges B & C', rev_id='D-id')
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='F-id')
+        self.expectFailure("Merge3Merger doesn't use lcas for symlink content",
+            self.assertEqual, 0, conflicts)
+        self.assertEqual('bing', wt.get_symlink_target('foo-id'))
+
+    def test_renamed_symlink(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A       Create symlink foo => bar
+        #  / \
+        # B   C     B renames foo => barry
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     D & E have barry
+        #     |
+        #     F     F renames barry to blah
+        #
+        # Merging D & F should result in F cleanly overriding D, because D's
+        # value actually comes from B
+
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('A add symlink', rev_id='A-id')
+        wt.rename_one('foo', 'barry')
+        wt.commit('B foo => barry', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('barry', wt.id2path('foo-id'))
+        self.assertEqual('bar', wt.get_symlink_target('foo-id'))
+        wt.commit('E merges C & B', rev_id='E-id')
+        wt.rename_one('barry', 'blah')
+        wt.commit('F barry => blah', rev_id='F-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        wt.commit('D merges B & C', rev_id='D-id')
+        self.assertEqual('barry', wt.id2path('foo-id'))
+        # Check the output of the Merger object directly
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'F-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        root_id = wt.path2id('')
+        entries = list(merge_obj._entries_lca())
+        # No content change, just a path change
+        self.assertEqual([('foo-id', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'barry', u'foo']), u'blah', u'barry'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='F-id')
+        self.assertEqual(0, conflicts)
+        self.assertEqual('blah', wt.id2path('foo-id'))
+
+    def test_symlink_no_content_change(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A       Create symlink foo => bar
+        #  / \
+        # B   C     B relinks foo => baz
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     D & E have foo => baz
+        # |
+        # F         F has foo => bing
+        #
+        # Merging E into F should not cause a conflict, because E doesn't have
+        # a content change relative to the LCAs (it does relative to A)
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('add symlink', rev_id='A-id')
+        os.remove('path/foo')
+        os.symlink('baz', 'path/foo')
+        wt.commit('foo => baz', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('baz', wt.get_symlink_target('foo-id'))
+        wt.commit('E merges C & B', rev_id='E-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        wt.commit('D merges B & C', rev_id='D-id')
+        os.remove('path/foo')
+        os.symlink('bing', 'path/foo')
+        wt.commit('F foo => bing', rev_id='F-id')
+
+        # Check the output of the Merger object directly
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        # Nothing interesting happened in OTHER relative to BASE
+        self.assertEqual([], list(merge_obj._entries_lca()))
+        # Now do a real merge, just to test the rest of the stack
+        conflicts = wt.merge_from_branch(wt.branch, to_revision='E-id')
+        self.assertEqual(0, conflicts)
+        self.assertEqual('bing', wt.get_symlink_target('foo-id'))
+
+    def test_symlink_this_changed_kind(self):
+        self.requireFeature(tests.SymlinkFeature)
+        #   A       Nothing
+        #  / \
+        # B   C     B creates symlink foo => bar
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     D changes foo into a file, E has foo => bing
+        #
+        # Mostly, this is trying to test that we don't try to os.readlink() on
+        # a file, or when there is nothing there
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        wt.commit('base', rev_id='A-id')
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('add symlink foo => bar', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('bar', wt.get_symlink_target('foo-id'))
+        os.remove('path/foo')
+        # We have to change the link in E, or it won't try to do a comparison
+        os.symlink('bing', 'path/foo')
+        wt.commit('E merges C & B, overrides to bing', rev_id='E-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        os.remove('path/foo')
+        self.build_tree_contents([('path/foo', 'file content\n')])
+        # XXX: workaround, WT doesn't detect kind changes unless you do
+        # iter_changes()
+        list(wt.iter_changes(wt.basis_tree()))
+        wt.commit('D merges B & C, makes it a file', rev_id='D-id')
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        root_id = wt.path2id('')
+        self.assertEqual([('foo-id', True,
+                           ((None, [root_id, None]), root_id, root_id),
+                           ((None, [u'foo', None]), u'foo', u'foo'),
+                           ((None, [False, None]), False, False)),
+                         ], entries)
+
+    def test_symlink_all_wt(self):
+        """Check behavior if all trees are Working Trees."""
+        self.requireFeature(tests.SymlinkFeature)
+        # The big issue is that entry.symlink_target is None for WorkingTrees.
+        # So we need to make sure we handle that case correctly.
+        #   A   foo => bar
+        #   |\
+        #   B C B relinks foo => baz
+        #   |X|
+        #   D E D & E have foo => baz
+        #     |
+        #     F F changes it to bing
+        # Merging D & F should result in F cleanly overriding D, because D's
+        # value actually comes from B
+
+        wt = self.make_branch_and_tree('path')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        os.symlink('bar', 'path/foo')
+        wt.add(['foo'], ['foo-id'])
+        wt.commit('add symlink', rev_id='A-id')
+        os.remove('path/foo')
+        os.symlink('baz', 'path/foo')
+        wt.commit('foo => baz', rev_id='B-id')
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.revert()
+        wt.commit('C', rev_id='C-id')
+        wt.merge_from_branch(wt.branch, 'B-id')
+        self.assertEqual('baz', wt.get_symlink_target('foo-id'))
+        wt.commit('E merges C & B', rev_id='E-id')
+        os.remove('path/foo')
+        os.symlink('bing', 'path/foo')
+        wt.commit('F foo => bing', rev_id='F-id')
+        wt.set_last_revision('B-id')
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.revert()
+        wt.merge_from_branch(wt.branch, 'C-id')
+        wt.commit('D merges B & C', rev_id='D-id')
+        wt_base = wt.bzrdir.sprout('base', 'A-id').open_workingtree()
+        wt_base.lock_read()
+        self.addCleanup(wt_base.unlock)
+        wt_lca1 = wt.bzrdir.sprout('b-tree', 'B-id').open_workingtree()
+        wt_lca1.lock_read()
+        self.addCleanup(wt_lca1.unlock)
+        wt_lca2 = wt.bzrdir.sprout('c-tree', 'C-id').open_workingtree()
+        wt_lca2.lock_read()
+        self.addCleanup(wt_lca2.unlock)
+        wt_other = wt.bzrdir.sprout('other', 'F-id').open_workingtree()
+        wt_other.lock_read()
+        self.addCleanup(wt_other.unlock)
+        merge_obj = _mod_merge.Merge3Merger(wt, wt, wt_base,
+            wt_other, lca_trees=[wt_lca1, wt_lca2], do_merge=False)
+        entries = list(merge_obj._entries_lca())
+        root_id = wt.path2id('')
+        self.assertEqual([('foo-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'foo', u'foo']), u'foo', u'foo'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_other_reverted_path_to_base(self):
+        #   A       Path at 'foo'
+        #  / \
+        # B   C     Path at 'bar' in B
+        # |\ /|
+        # | X |
+        # |/ \|
+        # D   E     Path at 'bar'
+        #     |
+        #     F     Path at 'foo'
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'a\nb\nc\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('rename', ('foo', 'bar'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('rename', ('foo', 'bar'))]) # merge the rename
+        builder.build_snapshot('F-id', ['E-id'],
+            [('rename', ('bar', 'foo'))]) # Rename back to BASE
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.assertEqual(0, conflicts)
+        self.assertEqual('foo', wt.id2path('foo-id'))
+
+    def test_other_reverted_content_to_base(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'base content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('foo-id', 'B content\n'))]) # merge the content
+        builder.build_snapshot('F-id', ['E-id'],
+            [('modify', ('foo-id', 'base content\n'))]) # Revert back to BASE
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.assertEqual(0, conflicts)
+        # TODO: We need to use the per-file graph to properly select a BASE
+        #       before this will work. Or at least use the LCA trees to find
+        #       the appropriate content base. (which is B, not A).
+        self.expectFailure("Merge3Merger doesn't recognize reverted content",
+            self.assertEqual, 'base content\n', wt.get_file_text('foo-id'))
+
+    def test_other_modified_content(self):
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'foo', 'foo-id', 'file', 'base content\n'))])
+        builder.build_snapshot('C-id', ['A-id'], [])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('modify', ('foo-id', 'B content\n'))]) # merge the content
+        builder.build_snapshot('F-id', ['E-id'],
+            [('modify', ('foo-id', 'F content\n'))]) # Override B content
+        builder.build_snapshot('D-id', ['B-id', 'C-id'], [])
+        wt, conflicts = self.do_merge(builder, 'F-id')
+        self.expectFailure("Merge3Merger only uses BASE for content",
+            self.assertEqual, 'F content\n', wt.get_file_text('foo-id'))
+        self.assertEqual(0, conflicts)
+        self.assertEqual('F content\n', wt.get_file_text('foo-id'))
+
+    def test_all_wt(self):
+        """Check behavior if all trees are Working Trees."""
+        # The big issue is that entry.revision is None for WorkingTrees. (as is
+        # entry.text_sha1, etc. So we need to make sure we handle that case
+        # correctly.
+        #   A   Content of 'foo', path of 'a'
+        #   |\
+        #   B C B modifies content, C renames 'a' => 'b'
+        #   |X|
+        #   D E E updates content, renames 'b' => 'c'
+        builder = self.get_builder()
+        builder.build_snapshot('A-id', None,
+            [('add', (u'', 'a-root-id', 'directory', None)),
+             ('add', (u'a', 'a-id', 'file', 'base content\n')),
+             ('add', (u'foo', 'foo-id', 'file', 'base content\n'))])
+        builder.build_snapshot('B-id', ['A-id'],
+            [('modify', ('foo-id', 'B content\n'))])
+        builder.build_snapshot('C-id', ['A-id'],
+            [('rename', ('a', 'b'))])
+        builder.build_snapshot('E-id', ['C-id', 'B-id'],
+            [('rename', ('b', 'c')),
+             ('modify', ('foo-id', 'E content\n'))])
+        builder.build_snapshot('D-id', ['B-id', 'C-id'],
+            [('rename', ('a', 'b'))]) # merged change
+        wt_this = self.get_wt_from_builder(builder)
+        wt_base = wt_this.bzrdir.sprout('base', 'A-id').open_workingtree()
+        wt_base.lock_read()
+        self.addCleanup(wt_base.unlock)
+        wt_lca1 = wt_this.bzrdir.sprout('b-tree', 'B-id').open_workingtree()
+        wt_lca1.lock_read()
+        self.addCleanup(wt_lca1.unlock)
+        wt_lca2 = wt_this.bzrdir.sprout('c-tree', 'C-id').open_workingtree()
+        wt_lca2.lock_read()
+        self.addCleanup(wt_lca2.unlock)
+        wt_other = wt_this.bzrdir.sprout('other', 'E-id').open_workingtree()
+        wt_other.lock_read()
+        self.addCleanup(wt_other.unlock)
+        merge_obj = _mod_merge.Merge3Merger(wt_this, wt_this, wt_base,
+            wt_other, lca_trees=[wt_lca1, wt_lca2], do_merge=False)
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('a-id', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'a', [u'a', u'b']), u'c', u'b'),
+                           ((False, [False, False]), False, False)),
+                          ('foo-id', True,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'foo', [u'foo', u'foo']), u'foo', u'foo'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_nested_tree_unmodified(self):
+        # Tested with a real WT, because BranchBuilder/MemoryTree don't handle
+        # 'tree-reference'
+        wt = self.make_branch_and_tree('tree',
+            format='dirstate-with-subtree')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        sub_tree = self.make_branch_and_tree('tree/sub-tree',
+            format='dirstate-with-subtree')
+        wt.set_root_id('a-root-id')
+        sub_tree.set_root_id('sub-tree-root')
+        self.build_tree_contents([('tree/sub-tree/file', 'text1')])
+        sub_tree.add('file')
+        sub_tree.commit('foo', rev_id='sub-A-id')
+        wt.add_reference(sub_tree)
+        wt.commit('set text to 1', rev_id='A-id', recursive=None)
+        # Now create a criss-cross merge in the parent, without modifying the
+        # subtree
+        wt.commit('B', rev_id='B-id', recursive=None)
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.commit('C', rev_id='C-id', recursive=None)
+        wt.merge_from_branch(wt.branch, to_revision='B-id')
+        wt.commit('E', rev_id='E-id', recursive=None)
+        wt.set_parent_ids(['B-id', 'C-id'])
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.commit('D', rev_id='D-id', recursive=None)
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        self.assertEqual([], entries)
+
+    def test_nested_tree_subtree_modified(self):
+        # Tested with a real WT, because BranchBuilder/MemoryTree don't handle
+        # 'tree-reference'
+        wt = self.make_branch_and_tree('tree',
+            format='dirstate-with-subtree')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        sub_tree = self.make_branch_and_tree('tree/sub',
+            format='dirstate-with-subtree')
+        wt.set_root_id('a-root-id')
+        sub_tree.set_root_id('sub-tree-root')
+        self.build_tree_contents([('tree/sub/file', 'text1')])
+        sub_tree.add('file')
+        sub_tree.commit('foo', rev_id='sub-A-id')
+        wt.add_reference(sub_tree)
+        wt.commit('set text to 1', rev_id='A-id', recursive=None)
+        # Now create a criss-cross merge in the parent, without modifying the
+        # subtree
+        wt.commit('B', rev_id='B-id', recursive=None)
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.commit('C', rev_id='C-id', recursive=None)
+        wt.merge_from_branch(wt.branch, to_revision='B-id')
+        self.build_tree_contents([('tree/sub/file', 'text2')])
+        sub_tree.commit('modify contents', rev_id='sub-B-id')
+        wt.commit('E', rev_id='E-id', recursive=None)
+        wt.set_parent_ids(['B-id', 'C-id'])
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.commit('D', rev_id='D-id', recursive=None)
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        # Nothing interesting about this sub-tree, because content changes are
+        # computed at a higher level
+        self.assertEqual([], entries)
+
+    def test_nested_tree_subtree_renamed(self):
+        # Tested with a real WT, because BranchBuilder/MemoryTree don't handle
+        # 'tree-reference'
+        wt = self.make_branch_and_tree('tree',
+            format='dirstate-with-subtree')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        sub_tree = self.make_branch_and_tree('tree/sub',
+            format='dirstate-with-subtree')
+        wt.set_root_id('a-root-id')
+        sub_tree.set_root_id('sub-tree-root')
+        self.build_tree_contents([('tree/sub/file', 'text1')])
+        sub_tree.add('file')
+        sub_tree.commit('foo', rev_id='sub-A-id')
+        wt.add_reference(sub_tree)
+        wt.commit('set text to 1', rev_id='A-id', recursive=None)
+        # Now create a criss-cross merge in the parent, without modifying the
+        # subtree
+        wt.commit('B', rev_id='B-id', recursive=None)
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.commit('C', rev_id='C-id', recursive=None)
+        wt.merge_from_branch(wt.branch, to_revision='B-id')
+        wt.rename_one('sub', 'alt_sub')
+        wt.commit('E', rev_id='E-id', recursive=None)
+        wt.set_last_revision('B-id')
+        wt.revert()
+        wt.set_parent_ids(['B-id', 'C-id'])
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.commit('D', rev_id='D-id', recursive=None)
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('sub-tree-root', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'sub', [u'sub', u'sub']), u'alt_sub', u'sub'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+    def test_nested_tree_subtree_renamed_and_modified(self):
+        # Tested with a real WT, because BranchBuilder/MemoryTree don't handle
+        # 'tree-reference'
+        wt = self.make_branch_and_tree('tree',
+            format='dirstate-with-subtree')
+        wt.lock_write()
+        self.addCleanup(wt.unlock)
+        sub_tree = self.make_branch_and_tree('tree/sub',
+            format='dirstate-with-subtree')
+        wt.set_root_id('a-root-id')
+        sub_tree.set_root_id('sub-tree-root')
+        self.build_tree_contents([('tree/sub/file', 'text1')])
+        sub_tree.add('file')
+        sub_tree.commit('foo', rev_id='sub-A-id')
+        wt.add_reference(sub_tree)
+        wt.commit('set text to 1', rev_id='A-id', recursive=None)
+        # Now create a criss-cross merge in the parent, without modifying the
+        # subtree
+        wt.commit('B', rev_id='B-id', recursive=None)
+        wt.set_last_revision('A-id')
+        wt.branch.set_last_revision_info(1, 'A-id')
+        wt.commit('C', rev_id='C-id', recursive=None)
+        wt.merge_from_branch(wt.branch, to_revision='B-id')
+        self.build_tree_contents([('tree/sub/file', 'text2')])
+        sub_tree.commit('modify contents', rev_id='sub-B-id')
+        wt.rename_one('sub', 'alt_sub')
+        wt.commit('E', rev_id='E-id', recursive=None)
+        wt.set_last_revision('B-id')
+        wt.revert()
+        wt.set_parent_ids(['B-id', 'C-id'])
+        wt.branch.set_last_revision_info(2, 'B-id')
+        wt.commit('D', rev_id='D-id', recursive=None)
+
+        merger = _mod_merge.Merger.from_revision_ids(progress.DummyProgress(),
+            wt, 'E-id')
+        merger.merge_type = _mod_merge.Merge3Merger
+        merge_obj = merger.make_merger()
+        entries = list(merge_obj._entries_lca())
+        root_id = 'a-root-id'
+        self.assertEqual([('sub-tree-root', False,
+                           ((root_id, [root_id, root_id]), root_id, root_id),
+                           ((u'sub', [u'sub', u'sub']), u'alt_sub', u'sub'),
+                           ((False, [False, False]), False, False)),
+                         ], entries)
+
+
+class TestLCAMultiWay(tests.TestCase):
+
+    def assertLCAMultiWay(self, expected, base, lcas, other, this,
+                          allow_overriding_lca=True):
+        self.assertEqual(expected, _mod_merge.Merge3Merger._lca_multi_way(
+                                (base, lcas), other, this,
+                                allow_overriding_lca=allow_overriding_lca))
+
+    def test_other_equal_equal_lcas(self):
+        """Test when OTHER=LCA and all LCAs are identical."""
+        self.assertLCAMultiWay('this',
+            'bval', ['bval', 'bval'], 'bval', 'bval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval'], 'lcaval', 'bval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval', 'lcaval'], 'lcaval', 'bval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval', 'lcaval'], 'lcaval', 'tval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval', 'lcaval'], 'lcaval', None)
+
+    def test_other_equal_this(self):
+        """Test when other and this are identical."""
+        self.assertLCAMultiWay('this',
+            'bval', ['bval', 'bval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lcaval', 'lcaval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            'bval', ['cval', 'dval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            'bval', [None, 'lcaval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            None, [None, 'lcaval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            None, ['lcaval', 'lcaval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            None, ['cval', 'dval'], 'oval', 'oval')
+        self.assertLCAMultiWay('this',
+            None, ['cval', 'dval'], None, None)
+        self.assertLCAMultiWay('this',
+            None, ['cval', 'dval', 'eval', 'fval'], 'oval', 'oval')
+
+    def test_no_lcas(self):
+        self.assertLCAMultiWay('this',
+            'bval', [], 'bval', 'tval')
+        self.assertLCAMultiWay('other',
+            'bval', [], 'oval', 'bval')
+        self.assertLCAMultiWay('conflict',
+            'bval', [], 'oval', 'tval')
+        self.assertLCAMultiWay('this',
+            'bval', [], 'oval', 'oval')
+
+    def test_lca_supersedes_other_lca(self):
+        """If one lca == base, the other lca takes precedence"""
+        self.assertLCAMultiWay('this',
+            'bval', ['bval', 'lcaval'], 'lcaval', 'tval')
+        self.assertLCAMultiWay('this',
+            'bval', ['bval', 'lcaval'], 'lcaval', 'bval')
+        # This is actually considered a 'revert' because the 'lcaval' in LCAS
+        # supersedes the BASE val (in the other LCA) but then OTHER reverts it
+        # back to bval.
+        self.assertLCAMultiWay('other',
+            'bval', ['bval', 'lcaval'], 'bval', 'lcaval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['bval', 'lcaval'], 'bval', 'tval')
+
+    def test_other_and_this_pick_different_lca(self):
+        # OTHER and THIS resolve the lca conflict in different ways
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'lca1val', 'lca2val')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'lca2val')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'bval'], 'lca1val', 'lca2val')
+
+    def test_other_in_lca(self):
+        # OTHER takes a value of one of the LCAs, THIS takes a new value, which
+        # theoretically supersedes both LCA values and 'wins'
+        self.assertLCAMultiWay('this',
+            'bval', ['lca1val', 'lca2val'], 'lca1val', 'newval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'newval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'lca1val', 'newval',
+            allow_overriding_lca=False)
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'newval',
+            allow_overriding_lca=False)
+        # THIS reverted back to BASE, but that is an explicit supersede of all
+        # LCAs
+        self.assertLCAMultiWay('this',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'bval')
+        self.assertLCAMultiWay('this',
+            'bval', ['lca1val', 'lca2val', 'bval'], 'lca1val', 'bval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'lca1val', 'bval',
+            allow_overriding_lca=False)
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'bval'], 'lca1val', 'bval',
+            allow_overriding_lca=False)
+
+    def test_this_in_lca(self):
+        # THIS takes a value of one of the LCAs, OTHER takes a new value, which
+        # theoretically supersedes both LCA values and 'wins'
+        self.assertLCAMultiWay('other',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'lca1val')
+        self.assertLCAMultiWay('other',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'lca2val')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'lca1val',
+            allow_overriding_lca=False)
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'lca2val',
+            allow_overriding_lca=False)
+        # OTHER reverted back to BASE, but that is an explicit supersede of all
+        # LCAs
+        self.assertLCAMultiWay('other',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'bval', 'lca3val')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'bval', 'lca3val',
+            allow_overriding_lca=False)
+
+    def test_all_differ(self):
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val'], 'oval', 'tval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca2val'], 'oval', 'tval')
+        self.assertLCAMultiWay('conflict',
+            'bval', ['lca1val', 'lca2val', 'lca3val'], 'oval', 'tval')

=== added file 'doc/developers/lca_tree_merging.txt'
--- doc/developers/lca_tree_merging.txt	1970-01-01 00:00:00 +0000
+++ doc/developers/lca_tree_merging.txt	2008-09-05 02:29:34 +0000
@@ -0,0 +1,126 @@
+================
+LCA Tree Merging
+================
+
+There are 2 ways that you get LCA merge resolution in bzr. First, if you use
+``bzr merge --lca``, the *content* of files will be resolved using a Least Common
+Ancestors algorithm. That is described in <lca-merge.html> not here.
+
+This document describes how we handle merging tree-shape when there is not
+a single unique ancestor (criss-cross merge). With a single LCA, we use
+simple 3-way-merge logic.
+
+When there are multiple possible LCAs, we use a different algorithm for
+handling tree-shape merging. Described here.
+
+As a simple example, here is a revision graph which we will refer to often::
+
+  .    BASE
+  .  /      \
+  . LCA1   LCA2
+  . |   \ /   |
+  . |    X    |
+  . |   / \   |
+  . THIS  OTHER
+
+In this graph, ``THIS`` and ``OTHER`` both have ``LCA1`` and ``LCA2`` in their
+ancestry but neither is an ancestor of the other, so we have 2 least common
+ancestors. The unique common ancestor is ``BASE``. (It should be noted that in
+this text we will talk directly about ``LCA1`` and ``LCA2``, but the algorithms
+are designed to cope with more than 2 LCAs.)
+
+
+Scalars
+=======
+
+Definition
+----------
+
+I'm defining scalar values as ones that cannot be 'merged' on their own. For
+example, the name of a file is "scalar". If one person changes "foo.txt" to
+"foo.c" and someone else changes "foo.txt" to "bar.txt" we don't merge the
+changes to be "bar.c", we simply conflict and expect the user to sort it out.
+
+We use a slightly different algorithm for scalars.
+
+
+Resolution Algorithm
+--------------------
+
+(This can be seen as ``bzrlib.merge.Merge3Merger._lca_multi_way```
+
+1. If ``THIS`` and ``OTHER`` have the same value, use it. There is no need to
+   inspect any other values in this case. Either nothing was changed (all
+   interesting nodes would have the same value), or we have "accidental
+   convergence" (both sides made the same change.).
+
+2. Find the values from ``LCA1`` and ``LCA2`` which are not the same as
+   ``BASE``. The idea here is to provide a rudimentary "heads" comparison.
+   Often, the whole tree graph will have a criss-cross, but the per-file
+   (per-scalar) graph would be linear, and the value in one LCA strictly
+   dominates the other. It is possible to construct a scenario where one side
+   dominates the other, but the dominated value is not ``BASE``, but a second
+   intermediate value. Most scalars are rarely changed, so this is unlikely to
+   be an issue. The trade-off is having to generate and inspect the
+   per-scalar graph.
+
+   If there are no LCA values that are different from ``BASE``, we use a simple
+   3-way merge with ``BASE`` as the base value.
+
+3. Find the unique set of LCA values that do not include the ``BASE`` value.
+   If there is only one unique LCA value, we again use three-way merge logic
+   using that unique value as the base.
+
+4. At this point, we have determined that we have at least 2 unique values in
+   our LCAs which means that ``THIS`` and ``OTHER`` would both have to resolve
+   the conflict. If they resolved it in the same way, we would have caught that
+   in step 1. So they either both picked a different LCA value, or one (or
+   both) chose a new value to use.
+
+   If ``OTHER`` and ``THIS`` both picked a different LCA value, we conflict.
+
+   If ``OTHER`` and ``THIS`` both have values that are not LCA values, we also
+   conflict. (Same as 3-way, both sides modified a value in different ways.)
+
+5. (optional) The only tricky part is this: if ``OTHER`` has a LCA value, but
+   ``THIS`` does not, then we go with ``THIS``, and conversely if ``THIS`` has
+   an LCA value, but ``OTHER`` does not, then we go with ``OTHER``. The idea is
+   that ``THIS`` and ``OTHER`` may have resolved things in the same way, and
+   then later changed the value to something newer. (They could have also
+   resolved it differently, and then one side updated again.)
+
+
+``InventoryEntry.revision``
+---------------------------
+
+The last-modified revision for an entry gets treated differently. This is
+because how it is generated allows us to infer more information. Specifically,
+any time there is a change to an entry (rename, or content change) the last
+modified revision is updated. Further, if we are merging, and both sides
+updated the entry, then we update the last-modified revision at the merge
+point.
+
+For a picture example::
+
+    .   A
+    .  / \
+    . B   C
+    .  \ /
+    .   D
+
+
+For a single entry, the last modified revision in ``D`` is:
+
+1) ``A`` if neither ``B`` or ``C`` modified it
+2) ``B`` if ``B`` modified and ``C`` did not
+3) ``C`` if ``C`` modified and ``B`` did not
+4) ``D`` if ``B`` and ``C`` modified it
+
+This means that if the last modified revision is the same, there have been no
+changes in the intermediate time. If ``OTHER`` also has the same last modified
+revision as *any* LCA, then we know that all other LCAs' last-modified
+revisions are in the ancestry of that value. (Otherwise, when ``OTHER`` would
+need to create a new last modified revision as part of the merge.)
+
+..
+   vim: ft=rst tw=74 ai

